# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2024 Christian Buhtz <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Tests related to Remove & Retention, formally known as Auto- and
Smart-remove.

About the current state of this test module:

    Most of the tests in this module are pseudo-tests. The do not test
    productive code but surrogates (e.g. method name `_org()` in each class).
    Because the productive code is in an untestable state and needs refactoring
    or totally rewrite. This is on the projects todo list. See meta issue
    #1945.
"""
import os
import sys
import inspect
from unittest import mock
from typing import Union
from datetime import date, time, datetime, timedelta
from pathlib import Path
from tempfile import TemporaryDirectory
import pyfakefs.fake_filesystem_unittest as pyfakefs_ut
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import config  # noqa: E402,RUF100
import snapshots  # noqa: E402,RUF100


def dt2sidstr(d: Union[date, datetime], t: time = None, tag: int = 123):
    """Create a SID identification string out ouf a date and time infos."""
    if not t:
        try:
            # If d is datetime
            t = d.time()
        except AttributeError:
            t = time(7, 42, 31)

    return datetime.combine(d, t).strftime(f'%Y%m%d-%H%M%S-{tag}')


def dt2str(d: Union[date, datetime]):
    return d.strftime('%a %d %b %Y')


def sid2str(sid):
    """Convert a SID string into human readable date incl. weekday."""

    if isinstance(sid, snapshots.SID):
        sid = str(sid)

    result = datetime.strptime(sid.split('-')[0], '%Y%m%d') \
        .date().strftime('%c').strip()

    if result.endswith(' 00:00:00'):
        result = result[:-9]

    return result


def create_SIDs(start_date: Union[date, datetime, list[date]],
                days: int,
                cfg: config.Config):
    sids = []

    if isinstance(start_date, list):
        the_dates = start_date
    else:
        the_dates = [start_date + timedelta(days=x) for x in range(days)]

    for d in the_dates:
        sids.append(snapshots.SID(dt2sidstr(d), cfg))

    return sorted(sids, reverse=True)


class KeepFirst(pyfakefs_ut.TestCase):
    """Test Snapshot.removeKeepFirst().

    PyFakeFS is used here because of Config file dependency."""

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def test_one_but_set(self):
        """Return value is always a set with always only one element."""
        # One SID for each of 20 days beginning with 5th March 2022 07:42:31
        sids = create_SIDs(
            datetime(2020, 3, 5, 7, 42, 31), 700, self.cfg)

        sut = self.sn.smartRemoveKeepFirst(
            sids, date(2021, 8, 5), datetime.now().date())

        self.assertIsInstance(sut, set)
        self.assertTrue(len(sut), 1)

    def test_simple_one(self):
        """First element in a range of SIDs"""
        sids = create_SIDs(
            datetime(2022, 3, 5, 7, 42, 31), 20, self.cfg)

        sut = self.sn.smartRemoveKeepFirst(
            sids, date(2022, 3, 5), datetime.now().date())

        sut = sut.pop()

        self.assertTrue(str(sut).startswith('20220324-074231-'))

    def test_min_included_max_not(self):
        """Minimum date is included in range, but max date not"""
        sids = create_SIDs(
            [
                datetime(2022, 3, 5),
                datetime(2022, 3, 3),
            ],
            None,
            self.cfg)

        sut = self.sn.smartRemoveKeepFirst(
            snapshots=sids,
            min_date=date(2022, 3, 3),
            max_date=date(2022, 3, 5),
        )

        self.assertEqual(len(sut), 1)

        sut = sut.pop()

        # the min_date is included
        self.assertEqual(sut.date.date(), date(2022, 3, 3))

    def test_no_date_ordering(self):
        """Hit first in the list and ignoring its date ordering.

        The list of snapshots is not ordered anywhere."""
        sids = []
        # April, 2016...
        for timestamp_string in ['20160424-215134-123',  # …24th
                                 # This SID will hit because it is the first
                                 # in the range specified.
                                 '20160422-030324-123',  # …22th
                                 '20160422-020324-123',  # …22th
                                 '20160422-010324-123',  # …22th
                                 # This might be the earliest/first SID in the
                                 # date range specified but it is not the first
                                 # in the list. So it won't be hit.
                                 '20160421-013218-123',  # …21th
                                 '20160410-134327-123']:  # …10th
            sids.append(snapshots.SID(timestamp_string, self.cfg))

        sut = self.sn.smartRemoveKeepFirst(sids,
                                            date(2016, 4, 20),
                                            date(2016, 4, 23))

        self.assertEqual(str(sut.pop()), '20160422-030324-123')

    def test_keep_first_range_outside(self):
        """No SID inside the specified range"""
        sids = []
        # April, 2016...
        for timestamp_string in ['20160424-215134-123',  # …24th
                                 '20160422-030324-123',  # …22th
                                 '20160422-020324-123',  # …22th
                                 '20160422-010324-123',  # …22th
                                 '20160421-013218-123',  # …21th
                                 '20160410-134327-123']:  # …10th
            sids.append(snapshots.SID(timestamp_string, self.cfg))

        # Between 11th and 18th April
        sut = self.sn.smartRemoveKeepFirst(sids,
                                           date(2016, 4, 11),
                                           date(2016, 4, 18))

        # None will hit, because no SID in that range.
        self.assertEqual(len(sut), 0)

    @mock.patch.object(snapshots.SID, 'failed', new_callable=lambda: True)
    def test_all_invalid(self, _mock_failed):
        """All SIDS invalid (not healthy)"""
        sids = create_SIDs(
            datetime(2022, 3, 5, 7, 42, 31), 20, self.cfg)

        # By default healthy/invalid status is irrelevant
        sut = self.sn.smartRemoveKeepFirst(
            sids, date(2022, 3, 5), datetime.now().date())
        self.assertTrue(len(sut), 1)

        # Now make it relevant
        sut = self.sn.smartRemoveKeepFirst(
            sids, date(2022, 3, 5), datetime.now().date(),
            keep_healthy=True)
        self.assertTrue(len(sut), 0)

    @mock.patch.object(snapshots.SID, 'failed', new_callable=mock.PropertyMock)
    def test_ignore_unhealthy(self, mock_failed):
        # The second call to failed-property returns True
        mock_failed.side_effect = [False, True, False, False, False, False]
        sids = []
        for timestamp_string in ['20160424-215134-123',
                                 # could be hit, but is NOT healthy
                                 '20160422-030324-123',
                                 # hit this
                                 '20160422-020324-123',
                                 '20160422-010324-123',
                                 '20160421-013218-123',
                                 '20160410-134327-123']:
            sids.append(snapshots.SID(timestamp_string, self.cfg))

        # keep the first healthy snapshot
        sut = self.sn.smartRemoveKeepFirst(sids,
                                           date(2016, 4, 20),
                                           date(2016, 4, 23),
                                           keep_healthy=True)
        self.assertEqual(str(sut.pop()), '20160422-020324-123')


class KeepAllForLast(pyfakefs_ut.TestCase):
    """Test Snapshot.removeKeepAll().

    Keep all snapshots for the last N days.
    PyFakeFS is used here because of Config file dependency."""

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def _org(self, snapshots, now, days_to_keep):
        """Simulated production code. Refactoring is on the todo list."""

        keep_all = days_to_keep
        keep = self.sn.smartRemoveKeepAll(
            snapshots,
            now - timedelta(days=keep_all-1),
            now + timedelta(days=1))

        return sorted(keep, reverse=True)

    def test_border(self):
        """The dates used in the user manual example.

        Here the current (just running incomplete) day is contained in the
        calculation.
        """
        sids = create_SIDs([
            datetime(2025, 4, 17, 22, 0),
            datetime(2025, 4, 17, 18, 1),
            datetime(2025, 4, 17, 12, 0),
            datetime(2025, 4, 17, 4, 0),
            datetime(2025, 4, 16, 8, 30),
            datetime(2025, 4, 15, 16, 0),
            datetime(2025, 4, 15, 0, 0),
            datetime(2025, 4, 14, 23, 59),
            datetime(2025, 4, 14, 9, 0),
            ],
            None,
            self.cfg)

        sut = self._org(
            snapshots=sids,
            now=datetime(2025, 4, 17, 22, 00).date(),
            days_to_keep=2)

        self.assertEqual(sut[0].date, datetime(2025, 4, 17, 22, 0))
        self.assertEqual(sut[1].date, datetime(2025, 4, 17, 18, 1))
        self.assertEqual(sut[2].date, datetime(2025, 4, 17, 12, 0))
        self.assertEqual(sut[3].date, datetime(2025, 4, 17, 4, 0))
        self.assertEqual(sut[4].date, datetime(2025, 4, 16, 8, 30))

    def test_simple(self):
        """Simple"""
        # 10th to 25th

        sids = create_SIDs(datetime(2024, 2, 10), 15, self.cfg)

        # keep...
        sut = self.sn.smartRemoveKeepAll(
            sids,
            # ... from 12th ...
            date(2024, 2, 12),
            # ... to 19th.
            date(2024, 2, 20)
        )

        self.assertEqual(len(sut), 8)

        sut = sorted(sut)

        self.assertEqual(sut[0].date.date(), date(2024, 2, 12))
        self.assertEqual(sut[1].date.date(), date(2024, 2, 13))
        self.assertEqual(sut[2].date.date(), date(2024, 2, 14))
        self.assertEqual(sut[3].date.date(), date(2024, 2, 15))
        self.assertEqual(sut[4].date.date(), date(2024, 2, 16))
        self.assertEqual(sut[5].date.date(), date(2024, 2, 17))
        self.assertEqual(sut[6].date.date(), date(2024, 2, 18))
        self.assertEqual(sut[7].date.date(), date(2024, 2, 19))


class KeepOneForLastNDays(pyfakefs_ut.TestCase):
    """Covering the smart remove setting 'Keep the last snapshot of each day
    for the last N  days.'.

    That logic is implemented in 'Snapshots.smartRemoveList()' but not testable
    in isolation. So for a first shot we just duplicate that code in this
    tests (see self._org()).
    """

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def _org(self, now, n_days, snapshots):
        """Copied and slightly refactored from inside
        'Snapshots.smartRemoveList()'.
        """

        keep = set()
        d = now
        for _ in range(0, n_days):
            keep |= self.sn.smartRemoveKeepFirst(
                snapshots,
                d,
                d + timedelta(days=1),
                keep_healthy=True)
            d -= timedelta(days=1)

        return sorted(keep, reverse=True)

    def test_doc_example(self):
        sids = create_SIDs([
            datetime(2025, 4, 17, 22, 0),
            datetime(2025, 4, 17, 4, 0),
            datetime(2025, 4, 16, 8, 30),
            datetime(2025, 4, 15, 16, 0),
            datetime(2025, 4, 15, 0, 0),
            datetime(2025, 4, 14, 23, 59),
            datetime(2025, 4, 13, 19, 0),
            datetime(2025, 4, 13, 7, 0),
            datetime(2025, 4, 12, 18, 45),
            datetime(2025, 4, 12, 18, 5),
            datetime(2025, 4, 11, 9, 0),
            ],
            None,
            self.cfg)

        sut = self._org(
            now=date(2025, 4, 17),
            n_days=5,
            snapshots=sids)

        self.assertEqual(sut[0].date, datetime(2025, 4, 17, 22, 0))
        self.assertEqual(sut[1].date, datetime(2025, 4, 16, 8, 30))
        self.assertEqual(sut[2].date, datetime(2025, 4, 15, 16, 0))
        self.assertEqual(sut[3].date, datetime(2025, 4, 14, 23, 59))
        self.assertEqual(sut[4].date, datetime(2025, 4, 13, 19, 0))


class KeepOneForLastNWeeks(pyfakefs_ut.TestCase):
    """Covering the smart remove setting 'Keep the last snapshot for each week for the
    last N weeks'.

    That logic is implemented in 'Snapshots.smartRemoveList()' but not testable
    in isolation. So for a first shot we just duplicate that code in this
    tests (see self._org()).
    """

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def _org(self, now, n_weeks, snapshots, keep_healthy=True):
        """Keep one per week for the last n_weeks weeks.

        Copied and slightly refactored from inside
        'Snapshots.smartRemoveList()'.
        """
        keep = set()

        # Sunday ??? (Sonntag) of previous week
        idx_date = now - timedelta(days=now.weekday())

        for _ in range(0, n_weeks):

            min_date = idx_date
            max_date = idx_date + timedelta(days=7)

            keep |= self.sn.smartRemoveKeepFirst(
                snapshots,
                min_date,
                max_date,
                keep_healthy=keep_healthy)

            idx_date -= timedelta(days=7)

        return sorted(keep, reverse=True)

    def test_doc_example(self):
        """Example used in manual"""
        sids = create_SIDs(
            [
                # 5 Weeks, each 3 days
                datetime(2025, 4, 17, 22, 0),
                datetime(2025, 4, 16, 4, 0),
                datetime(2025, 4, 15, 14, 0),
                datetime(2025, 4, 13, 22, 0),
                datetime(2025, 4, 9, 4, 0),
                datetime(2025, 4, 8, 14, 0),
                datetime(2025, 4, 3, 22, 0),
                datetime(2025, 4, 2, 4, 0),
                datetime(2025, 4, 1, 14, 0),
                datetime(2025, 3, 27, 22, 0),
                datetime(2025, 3, 26, 4, 0),
                datetime(2025, 3, 24, 14, 0),
                datetime(2025, 3, 20, 22, 0),
                datetime(2025, 3, 19, 4, 0),
                datetime(2025, 3, 18, 14, 0)
            ],
            None,
            self.cfg
        )

        sut = self._org(
            now=date(2025, 4, 17),
            n_weeks=4,
            snapshots=sids)

        self.assertEqual(sut[0].date, datetime(2025, 4, 17, 22, 0))
        self.assertEqual(sut[1].date, datetime(2025, 4, 13, 22, 0))
        self.assertEqual(sut[2].date, datetime(2025, 4, 3, 22, 0))
        self.assertEqual(sut[3].date, datetime(2025, 3, 27, 22, 0))
        self.assertEqual(len(sut), 4)


class KeepOneForLastNMonths(pyfakefs_ut.TestCase):
    """Covering the smart remove setting 'Keep the last snapshot for each month
    for the last N months'.

    That logic is implemented in 'Snapshots.smartRemoveList()' but not testable
    in isolation. So for a first shot we just duplicate that code in this
    tests (see self._org()).

    """

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def _org(self, now, n_months, snapshots, keep_healthy=True):
        """Keep one per months for the last n_months months.

        Copied and slightly refactored from inside
        'Snapshots.smartRemoveList()'.
        """
        keep = set()

        d1 = date(now.year, now.month, 1)
        d2 = self.sn.incMonth(d1)

        # each months
        for _ in range(0, n_months):
            keep |= self.sn.smartRemoveKeepFirst(
                snapshots, d1, d2, keep_healthy=keep_healthy)
            d2 = d1
            d1 = self.sn.decMonth(d1)

        return sorted(keep, reverse=True)

    def test_doc_example(self):
        sids = create_SIDs(
            [
                # 10 months period
                date(2025, 8, 18),
                date(2025, 8, 6),
                date(2025, 7, 31),
                date(2025, 7, 1),
                date(2025, 6, 30),
                date(2025, 6, 1),
                # gap of 2 months
                date(2025, 3, 18),
                date(2025, 2, 9),
                date(2025, 1, 6),
                date(2024, 12, 26),
                date(2024, 11, 14),
            ],
            None,
            self.cfg
        )
        now = sids[0].date.date() + timedelta(days=5)

        months = 6
        sut = self._org(
            now=now,
            # Keep the last week
            n_months=months,
            snapshots=sids)

        expect = [
            date(2025, 8, 18),
            date(2025, 7, 31),
            date(2025, 6, 30),
            date(2025, 3, 18),
        ]
        self.assertEqual(len(sut), len(expect))
        for idx, expect_date in enumerate(expect):
            self.assertEqual(sut[idx].date.date(), expect_date)


class KeepOnePerYearForAllYears(pyfakefs_ut.TestCase):
    """Covering the smart remove setting 'Keep the last snapshot for each year
    for all years.'

    That logic is implemented in 'Snapshots.smartRemoveList()' but not testable
    in isolation. So for a first shot we just duplicate that code in this
    tests (see self._org()).
    """

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def _org(self, now, snapshots, keep_healthy=True):
        """Keep one per year

        Copied and slightly refactored from inside
        'Snapshots.smartRemoveList()'.
        """
        first_year = int(snapshots[-1].sid[:4])

        keep = set()

        for i in range(first_year, now.year+1):
            keep |= self.sn.smartRemoveKeepFirst(
                snapshots,
                date(i, 1, 1),
                date(i+1, 1, 1),
                keep_healthy=keep_healthy)

        return sorted(keep, reverse=True)

    def test_doc_example(self):
        now = date(2024, 12, 16)
        sids = create_SIDs(
            [
                date(2024, 10, 26),
                date(2024, 4, 13),
                date(2023, 10, 26),
                date(2023, 10, 8),
                date(2023, 1, 1),
                date(2022, 12, 31),
                date(2022, 4, 13),
                date(2020, 10, 26),
                date(2020, 4, 13),
            ],
            None,
            self.cfg
        )

        sut = self._org(
            now=now,
            snapshots=sids)

        expect = [
            date(2024, 10, 26),
            date(2023, 10, 26),
            date(2022, 12, 31),
            date(2020, 10, 26),
        ]
        self.assertEqual(len(sut), len(expect))
        for idx, expect_date in enumerate(expect):
            self.assertTrue(sut[idx].date.date(), expect_date)


class IncDecMonths(pyfakefs_ut.TestCase):
    """PyFakeFS is used here because of Config file dependency."""

    def setUp(self):
        """Setup a fake filesystem."""
        self.setUpPyfakefs(allow_root_user=False)

        # cleanup() happens automatically
        self._temp_dir = TemporaryDirectory(prefix='bit.')
        # Workaround: tempfile and pathlib not compatible yet
        self.temp_path = Path(self._temp_dir.name)

        self._config_fp = self._create_config_file(parent_path=self.temp_path)
        self.cfg = config.Config(str(self._config_fp))

        self.sn = snapshots.Snapshots(self.cfg)

    def _create_config_file(self, parent_path):
        """Minimal config file"""
        # pylint: disable-next=R0801
        cfg_content = inspect.cleandoc('''
            config.version=6
            profile1.snapshots.include.1.type=0
            profile1.snapshots.include.1.value=rootpath/source
            profile1.snapshots.include.size=1
            profile1.snapshots.no_on_battery=false
            profile1.snapshots.notify.enabled=true
            profile1.snapshots.path=rootpath/destination
            profile1.snapshots.path.host=test-host
            profile1.snapshots.path.profile=1
            profile1.snapshots.path.user=test-user
            profile1.snapshots.preserve_acl=false
            profile1.snapshots.preserve_xattr=false
            profile1.snapshots.remove_old_snapshots.enabled=true
            profile1.snapshots.remove_old_snapshots.unit=80
            profile1.snapshots.remove_old_snapshots.value=10
            profile1.snapshots.rsync_options.enabled=false
            profile1.snapshots.rsync_options.value=
            profiles.version=1
        ''')

        # config file location
        config_fp = parent_path / 'config_path' / 'config'
        config_fp.parent.mkdir()
        config_fp.write_text(cfg_content, 'utf-8')

        return config_fp

    def test_inc_simple(self):
        sut = self.sn.incMonth(date(1982, 8, 6))
        self.assertEqual(sut, date(1982, 9, 1))

    def test_inc_next_year(self):
        sut = self.sn.incMonth(date(1982, 12, 16))
        self.assertEqual(sut, date(1983, 1, 1))

    def test_inc_leap_year(self):
        sut = self.sn.incMonth(date(2020, 12, 16))
        self.assertEqual(sut, date(2021, 1, 1))

    def test_inc_leap_months(self):
        sut = self.sn.incMonth(date(2020, 2, 29))
        self.assertEqual(sut, date(2020, 3, 1))

    def test_dec_simple(self):
        sut = self.sn.decMonth(date(1982, 8, 6))
        self.assertEqual(sut, date(1982, 7, 1))

    def test_dec_year(self):
        sut = self.sn.decMonth(date(1982, 1, 6))
        self.assertEqual(sut, date(1981, 12, 1))

    def test_dec_leap_months(self):
        sut = self.sn.decMonth(date(2020, 2, 29))
        self.assertEqual(sut, date(2020, 1, 1))



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2008-2022 Taylor Raack
# SPDX-FileCopyrightText: © 2024 Christian Buhtz <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Tests about the tools module."""
import os
import sys
import subprocess
import random
import pathlib
import gzip
import stat
import signal
import unittest
from datetime import datetime
from time import sleep
from unittest.mock import patch
from copy import deepcopy
from tempfile import NamedTemporaryFile, TemporaryDirectory
import pyfakefs.fake_filesystem_unittest as pyfakefs_ut
from test import generic
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import tools
import configfile
from bitbase import TimeUnit

# chroot jails used for building may have no UUID devices (because of tmpfs)
# we need to skip tests that require UUIDs
DISK_BY_UUID_AVAILABLE = os.path.exists(tools.DISK_BY_UUID)

UDEVADM_HAS_UUID = subprocess.Popen(
    ['udevadm', 'info', '-e'],
    stdout=subprocess.PIPE,
    stderr=subprocess.DEVNULL).communicate()[0].find(b'ID_FS_UUID=') > 0

RSYNC_INSTALLED = tools.checkCommand('rsync')

RSYNC_307_VERSION = """rsync  version 3.0.7  protocol version 30
Copyright (C) 1996-2009 by Andrew Tridgell, Wayne Davison, and others.
Web site: http://rsync.samba.org/
Capabilities:
    64-bit files, 64-bit inums, 32-bit timestamps, 64-bit long ints,
    socketpairs, hardlinks, symlinks, IPv6, batchfiles, inplace,
    append, ACLs, xattrs, iconv, symtimes

rsync comes with ABSOLUTELY NO WARRANTY.  This is free software, and you
are welcome to redistribute it under certain conditions.  See the GNU
General Public License for details.
"""

RSYNC_310_VERSION = """rsync  version 3.1.0  protocol version 31
Copyright (C) 1996-2013 by Andrew Tridgell, Wayne Davison, and others.
Web site: http://rsync.samba.org/
Capabilities:
    64-bit files, 64-bit inums, 64-bit timestamps, 64-bit long ints,
    socketpairs, hardlinks, symlinks, IPv6, batchfiles, inplace,
    append, ACLs, xattrs, iconv, symtimes, prealloc

rsync comes with ABSOLUTELY NO WARRANTY.  This is free software, and you
are welcome to redistribute it under certain conditions.  See the GNU
General Public License for details.
"""


class TestTools(generic.TestCase):
    """
    All functions test here come from tools.py
    """

    def setUp(self):
        super().setUp()
        self.subproc = None

    def tearDown(self):
        super().tearDown()
        self._kill_process()

    def _create_process(self, *args):
        dummyPath = os.path.join(os.path.dirname(__file__), generic.DUMMY)
        cmd = [dummyPath]
        cmd.extend(args)
        self.subproc = subprocess.Popen(cmd)
        sleep(0.1)
        return self.subproc.pid

    def _kill_process(self):
        if self.subproc:
            self.subproc.kill()
            self.subproc.wait()
        self.subproc = None

    def test_sharePath(self):
        share = tools.sharePath()
        self.assertTrue(share.endswith('share'), 'share = {}'.format(share))

    def test_backintimePath(self):
        path = tools.backintimePath('common')
        self.assertIn(path, __file__)

    def test_registerBackintimePath(self):
        path = tools.backintimePath('foo')
        tools.registerBackintimePath('foo')
        self.assertIn(path, sys.path)
        sys.path.remove(path)

    def test_runningFromSource(self):
        self.assertTrue(tools.runningFromSource())

    def test_addSourceToPathEnviron(self):
        source = tools.backintimePath('common')
        path = [x for x in os.getenv('PATH').split(':') if x != source]
        os.environ['PATH'] = ':'.join(path)

        tools.addSourceToPathEnviron()
        self.assertIn(source, os.environ['PATH'])

    def test_readFile(self):
        """
        Test the function readFile
        """
        test_tools_file = os.path.abspath(__file__)
        test_directory = os.path.dirname(test_tools_file)
        non_existing_file = os.path.join(test_directory, "nonExistingFile")

        self.assertIsInstance(tools.readFile(test_tools_file), str)
        self.assertIsNone(tools.readFile(non_existing_file))

        with NamedTemporaryFile('wt') as tmp:
            tmp.write('foo\nbar')
            tmp.flush()
            self.assertIsInstance(tools.readFile(tmp.name), str)
            self.assertEqual(tools.readFile(tmp.name), 'foo\nbar')

        tmp_gz = NamedTemporaryFile().name
        with gzip.open(tmp_gz + '.gz', 'wt') as f:
            f.write('foo\nbar')
            f.flush()
        self.assertIsInstance(tools.readFile(tmp_gz), str)
        self.assertEqual(tools.readFile(tmp_gz), 'foo\nbar')
        os.remove(tmp_gz + '.gz')

    def test_readFileLines(self):
        """
        Test the function readFileLines
        """
        test_tools_file = os.path.abspath(__file__)
        test_directory = os.path.dirname(test_tools_file)
        non_existing_file = os.path.join(test_directory, "nonExistingFile")

        output = tools.readFileLines(test_tools_file)
        self.assertIsInstance(output, list)
        self.assertGreaterEqual(len(output), 1)
        self.assertIsInstance(output[0], str)
        self.assertIsNone(tools.readFileLines(non_existing_file))

        with NamedTemporaryFile('wt') as tmp:
            tmp.write('foo\nbar')
            tmp.flush()
            self.assertIsInstance(tools.readFileLines(tmp.name), list)
            self.assertListEqual(tools.readFileLines(tmp.name), ['foo', 'bar'])

        tmp_gz = NamedTemporaryFile().name
        with gzip.open(tmp_gz + '.gz', 'wt') as f:
            f.write('foo\nbar')
            f.flush()
        self.assertIsInstance(tools.readFileLines(tmp_gz), list)
        self.assertEqual(tools.readFileLines(tmp_gz), ['foo', 'bar'])
        os.remove(tmp_gz + '.gz')

    def test_checkCommand(self):
        """
        Test the function checkCommand
        """
        self.assertFalse(tools.checkCommand(''))
        self.assertFalse(tools.checkCommand("notExistedCommand"))
        self.assertTrue(tools.checkCommand("ls"))
        self.assertTrue(tools.checkCommand('backintime'))

    def test_which(self):
        """
        Test the function which
        """
        self.assertRegex(tools.which("ls"), r'/.*/ls')
        self.assertEqual(tools.which('backintime'),
                         os.path.join(os.getcwd(), 'backintime'))
        self.assertIsNone(tools.which("notExistedCommand"))

    def test_makeDirs(self):
        self.assertFalse(tools.makeDirs('/'))
        self.assertTrue(tools.makeDirs(os.getcwd()))
        with TemporaryDirectory() as d:
            path = os.path.join(d, 'foo', 'bar')
            self.assertTrue(tools.makeDirs(path))

    def test_makeDirs_not_writable(self):
        with TemporaryDirectory() as d:
            os.chmod(d, stat.S_IRUSR)
            path = os.path.join(
                d, 'foobar{}'.format(random.randrange(100, 999)))
            self.assertFalse(tools.makeDirs(path))

    def test_mkdir(self):
        self.assertFalse(tools.mkdir('/'))
        with TemporaryDirectory() as d:
            path = os.path.join(d, 'foo')
            self.assertTrue(tools.mkdir(path))
            for mode in (0o700, 0o644, 0o777):
                msg = 'new path should have octal permissions {0:#o}' \
                      .format(mode)
                path = os.path.join(d, '{0:#o}'.format(mode))
                self.assertTrue(tools.mkdir(path, mode), msg)
                self.assertEqual(
                    '{0:o}'.format(os.stat(path).st_mode & 0o777),
                    '{0:o}'.format(mode), msg)

    def test_pids(self):
        pids = tools.pids()
        self.assertGreater(len(pids), 0)
        self.assertIn(os.getpid(), pids)

    def test_processStat(self):
        pid = self._create_process()
        stat = tools.processStat(pid)
        self.assertRegex(
            stat, r'{} \({}\) \w .*'.format(pid, generic.DUMMY[:15]))

    @patch('builtins.open')
    def test_processStat_exception(self, mock_open):
        mock_open.side_effect = OSError()
        pid = self._create_process()
        self.assertEqual(tools.processStat(pid), '')

    def test_processPaused(self):
        pid = self._create_process()
        self.assertFalse(tools.processPaused(pid))
        self.subproc.send_signal(signal.SIGSTOP)
        sleep(0.01)
        self.assertTrue(tools.processPaused(pid))
        self.subproc.send_signal(signal.SIGCONT)
        sleep(0.01)
        self.assertFalse(tools.processPaused(pid))

    def test_processName(self):
        pid = self._create_process()
        self.assertEqual(tools.processName(pid), generic.DUMMY[:15])

    def test_processCmdline(self):
        pid = self._create_process()
        self.assertRegex(tools.processCmdline(pid),
                         r'.*/sh.*/common/test/dummy_test_process\.sh')
        self._kill_process()
        pid = self._create_process('foo', 'bar')
        self.assertRegex(tools.processCmdline(pid),
                         r'.*/sh.*/common/test/dummy_test_process\.sh.foo.bar')

    @patch('builtins.open')
    def test_processCmdline_exception(self, mock_open):
        mock_open.side_effect = OSError()
        pid = self._create_process()
        self.assertEqual(tools.processCmdline(pid), '')

    def test_pidsWithName(self):
        self.assertEqual(len(tools.pidsWithName('nonExistingProcess')), 0)
        pid = self._create_process()
        pids = tools.pidsWithName(generic.DUMMY)
        self.assertGreaterEqual(len(pids), 1)
        self.assertIn(pid, pids)

    def test_processExists(self):
        self.assertFalse(tools.processExists('nonExistingProcess'))
        self._create_process()
        self.assertTrue(tools.processExists(generic.DUMMY))

    def test_processAlive(self):
        """
        Test the function processAlive
        """
        # TODO: add test (in chroot) running proc as root and kill as non-root
        self.assertTrue(tools.processAlive(os.getpid()))
        pid = self._create_process()
        self.assertTrue(tools.processAlive(pid))
        self._kill_process()
        self.assertFalse(tools.processAlive(pid))
        self.assertFalse(tools.processAlive(999999))
        with self.assertRaises(ValueError):
            tools.processAlive(0)
        self.assertFalse(tools.processAlive(-1))

    def test_checkXServer(self):
        try:
            tools.checkXServer()
        except Exception as e:
            self.fail(
                'tools.ckeck_x_server() raised exception {}'.format(str(e)))

    def test_preparePath(self):
        path_with_slash_at_begin = "/test/path"
        path_without_slash_at_begin = "test/path"
        path_with_slash_at_end = "/test/path/"
        path_without_slash_at_end = "/test/path"
        self.assertEqual(
            tools.preparePath(path_with_slash_at_begin),
            path_with_slash_at_begin)
        self.assertEqual(
            tools.preparePath(path_without_slash_at_begin),
            path_with_slash_at_begin)
        self.assertEqual(
            tools.preparePath(path_without_slash_at_end),
            path_without_slash_at_end)
        self.assertEqual(
            tools.preparePath(path_with_slash_at_end),
            path_without_slash_at_end)

    def test_powerStatusAvailable(self):
        if tools.processExists('upowerd') and not generic.ON_TRAVIS:
            self.assertTrue(tools.powerStatusAvailable())
        else:
            self.assertFalse(tools.powerStatusAvailable())
        self.assertIsInstance(tools.onBattery(), bool)

    def test_rsyncCaps(self):
        if RSYNC_INSTALLED:
            caps = tools.rsyncCaps()
            self.assertIsInstance(caps, list)
            self.assertGreaterEqual(len(caps), 1)

        self.assertListEqual(tools.rsyncCaps(data=RSYNC_307_VERSION),
                             ['64-bit files',
                              '64-bit inums',
                              '32-bit timestamps',
                              '64-bit long ints',
                              'socketpairs',
                              'hardlinks',
                              'symlinks',
                              'IPv6',
                              'batchfiles',
                              'inplace',
                              'append',
                              'ACLs',
                              'xattrs',
                              'iconv',
                              'symtimes'])

        self.assertListEqual(tools.rsyncCaps(data=RSYNC_310_VERSION),
                             ['progress2',
                              '64-bit files',
                              '64-bit inums',
                              '64-bit timestamps',
                              '64-bit long ints',
                              'socketpairs',
                              'hardlinks',
                              'symlinks',
                              'IPv6',
                              'batchfiles',
                              'inplace',
                              'append',
                              'ACLs',
                              'xattrs',
                              'iconv',
                              'symtimes',
                              'prealloc'])

    def test_md5sum(self):
        with NamedTemporaryFile() as f:
            f.write(b'foo')
            f.flush()

            self.assertEqual(tools.md5sum(f.name),
                             'acbd18db4cc2f85cedef654fccc4a4d8')

    def test_checkCronPattern(self):
        self.assertTrue(tools.checkCronPattern('0'))
        self.assertTrue(tools.checkCronPattern('0,10,13,15,17,20,23'))
        self.assertTrue(tools.checkCronPattern('*/6'))
        self.assertFalse(tools.checkCronPattern('a'))
        self.assertFalse(tools.checkCronPattern(' 1'))
        self.assertFalse(tools.checkCronPattern('0,10,13,1a,17,20,23'))
        self.assertFalse(tools.checkCronPattern('0,10,13, 15,17,20,23'))
        self.assertFalse(tools.checkCronPattern('*/6,8'))
        self.assertFalse(tools.checkCronPattern('*/6 a'))

    def test_mountpoint(self):
        self.assertEqual(tools.mountpoint('/nonExistingFolder/foo/bar'), '/')
        proc = os.path.join('/proc', str(os.getpid()), 'fd')
        self.assertEqual(tools.mountpoint(proc), '/proc')

    def test_decodeOctalEscape(self):
        self.assertEqual(tools.decodeOctalEscape('/mnt/normalPath'),
                         '/mnt/normalPath')
        self.assertEqual(
            tools.decodeOctalEscape('/mnt/path\\040with\\040space'),
            '/mnt/path with space')

    def test_mountArgs(self):
        rootArgs = tools.mountArgs('/')
        self.assertIsInstance(rootArgs, list)
        self.assertGreaterEqual(len(rootArgs), 3)
        self.assertEqual(rootArgs[1], '/')

        procArgs = tools.mountArgs('/proc')
        self.assertGreaterEqual(len(procArgs), 3)
        self.assertEqual(procArgs[0], 'proc')
        self.assertEqual(procArgs[1], '/proc')
        self.assertEqual(procArgs[2], 'proc')

    def test_isRoot(self):
        self.assertIsInstance(tools.isRoot(), bool)

    def test_usingSudo(self):
        self.assertIsInstance(tools.usingSudo(), bool)

    def test_patternHasNotEncryptableWildcard(self):
        self.assertFalse(tools.patternHasNotEncryptableWildcard('foo'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('/foo'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('foo/*/bar'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('foo/**/bar'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('*/foo'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('**/foo'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('foo/*'))
        self.assertFalse(tools.patternHasNotEncryptableWildcard('foo/**'))

        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo?'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo[1-2]'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo*'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('*foo'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('**foo'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('*.foo'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo*bar'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo**bar'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo*/bar'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo**/bar'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo/*bar'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo/**bar'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('foo/*/bar*'))
        self.assertTrue(tools.patternHasNotEncryptableWildcard('*foo/*/bar'))

    def test_readTimeStamp(self):
        with NamedTemporaryFile('wt') as f:
            f.write('20160127 0124')
            f.flush()
            self.assertEqual(tools.readTimeStamp(f.name),
                             datetime(2016, 1, 27, 1, 24))

        with NamedTemporaryFile('wt') as f:
            f.write('20160127')
            f.flush()
            self.assertEqual(tools.readTimeStamp(f.name),
                             datetime(2016, 1, 27, 0, 0))

    def test_writeTimeStamp(self):
        with NamedTemporaryFile('rt') as f:
            tools.writeTimeStamp(f.name)
            s = f.read().strip('\n')
            self.assertTrue(s.replace(' ', '').isdigit())
            self.assertEqual(len(s), 13)

    def test_splitCommands(self):
        ret = list(tools.splitCommands(['echo foo;'],
                                       head='echo start;',
                                       tail='echo end',
                                       maxLength=40))
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0], 'echo start;echo foo;echo end')

        ret = list(tools.splitCommands(['echo foo;']*3,
                                       head='echo start;',
                                       tail='echo end',
                                       maxLength=40))
        self.assertEqual(len(ret), 2)
        self.assertEqual(ret[0], 'echo start;echo foo;echo foo;echo end')
        self.assertEqual(ret[1], 'echo start;echo foo;echo end')

        ret = list(tools.splitCommands(['echo foo;']*3,
                                       head='echo start;',
                                       tail='echo end',
                                       maxLength=0))
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0],
                         'echo start;echo foo;echo foo;echo foo;echo end')

        ret = list(tools.splitCommands(['echo foo;'] * 3,
                                       head='echo start;',
                                       tail='echo end',
                                       maxLength=-10))
        self.assertEqual(len(ret), 1)
        self.assertEqual(
            ret[0],
            'echo start;echo foo;echo foo;echo foo;echo end')


class EscapeIPv6(generic.TestCase):
    def test_escaped(self):
        values_and_expected = (
            ('fd00:0::5', '[fd00:0::5]'),
            (
                '2001:db8:0:8d3:0:8a2e:70:7344',
                '[2001:db8:0:8d3:0:8a2e:70:7344]'
            ),
            ('::', '[::]'),
            ('::1', '[::1]'),
            ('::ffff:192.0.2.128', '[::ffff:192.0.2.128]'),
            ('fe80::1', '[fe80::1]'),
        )

        for val, exp in values_and_expected:
            with self.subTest(val=val, exp=exp):
                self.assertEqual(tools.escapeIPv6Address(val), exp)

    def test_passed(self):
        test_values = (
            '192.168.1.1',
            '172.17.1.133',
            '255.255.255.255',
            '169.254.0.1',
        )

        for val in test_values:
            with self.subTest(val=val):
                # IPv4 addresses are not escaped
                self.assertEqual(tools.escapeIPv6Address(val), val)

    def test_invalid(self):
        """Invalid IP addresses and hostnames"""
        test_values = (
            'foo.bar',
            'fd00',
            '2001:0db8:::0000:0000:8a2e:0370:7334',
            ':2001:0db8:85a3:0000:0000:8a2e:0370:7334',
            '2001:0gb8:85a3:0000:0000:8a2e:0370:7334',
            '2001:0db8:85a3:0000:0000:8a2e:0370:7334:abcd',
            'localhost',
        )

        for val in test_values:
            with self.subTest(val=val):
                self.assertEqual(tools.escapeIPv6Address(val), val)


class Environ(generic.TestCase):
    """???
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.env = deepcopy(os.environ)

    def setUp(self):
        super().setUp()
        self.temp_file = '/tmp/temp.txt'
        os.environ = deepcopy(self.env)

    def tearDown(self):
        super().tearDown()
        if os.path.exists(self.temp_file):
            os.remove(self.temp_file)
        os.environ = deepcopy(self.env)

    def test_envLoad_without_previous_values(self):
        test_env = configfile.ConfigFile()
        test_env.setStrValue('FOO', 'bar')
        test_env.setStrValue('ASDF', 'qwertz')
        test_env.save(self.temp_file)

        # make sure environ is clean
        self.assertNotIn('FOO', os.environ)
        self.assertNotIn('ASDF', os.environ)

        tools.envLoad(self.temp_file)
        self.assertIn('FOO', os.environ)
        self.assertIn('ASDF', os.environ)
        self.assertEqual(os.environ['FOO'], 'bar')
        self.assertEqual(os.environ['ASDF'], 'qwertz')

    def test_envLoad_do_not_overwrite_previous_values(self):
        test_env = configfile.ConfigFile()
        test_env.setStrValue('FOO', 'bar')
        test_env.setStrValue('ASDF', 'qwertz')
        test_env.save(self.temp_file)

        # add some environ vars that should not get overwritten
        os.environ['FOO'] = 'defaultFOO'
        os.environ['ASDF'] = 'defaultASDF'

        tools.envLoad(self.temp_file)
        self.assertIn('FOO', os.environ)
        self.assertIn('ASDF', os.environ)
        self.assertEqual(os.environ['FOO'], 'defaultFOO')
        self.assertEqual(os.environ['ASDF'], 'defaultASDF')

    def test_envSave(self):
        keys = (
            'GNOME_KEYRING_CONTROL',
            'DBUS_SESSION_BUS_ADDRESS',
            'DBUS_SESSION_BUS_PID',
            'DBUS_SESSION_BUS_WINDOWID',
            'DISPLAY',
            'XAUTHORITY',
            'GNOME_DESKTOP_SESSION_ID',
            'KDE_FULL_SESSION')

        for i, k in enumerate(keys):
            os.environ[k] = str(i)

        tools.envSave(self.temp_file)

        self.assertIsFile(self.temp_file)

        test_env = configfile.ConfigFile()
        test_env.load(self.temp_file)
        for i, k in enumerate(keys):
            with self.subTest(i=i, k=k):
                # workaround for py.test3 2.5.1 doesn't support subTest
                msg = 'i = %s, k = %s' % (i, k)
                self.assertEqual(test_env.strValue(k), str(i), msg)


class ExecuteSubprocess(generic.TestCase):
    # new method with subprocess
    def test_returncode(self):
        self.assertEqual(tools.Execute(['true']).run(), 0)
        self.assertEqual(tools.Execute(['false']).run(), 1)

    def test_callback(self):
        c = lambda x, y: self.callback(self.assertEqual, x, 'foo')
        tools.Execute(['echo', 'foo'], callback=c).run()
        self.assertTrue(self.run)
        self.run = False

        # give extra user_data for callback
        c = lambda x, y: self.callback(self.assertEqual, x, y)
        tools.Execute(['echo', 'foo'], callback=c, user_data='foo').run()
        self.assertTrue(self.run)
        self.run = False

        # no output
        c = lambda x, y: self.callback(self.fail,
                                       'callback was called unexpectedly')
        tools.Execute(['true'], callback=c).run()
        self.assertFalse(self.run)
        self.run = False

    def test_pausable(self):
        proc = tools.Execute(['true'])
        self.assertTrue(proc.pausable)


class Tools_FakeFS(pyfakefs_ut.TestCase):
    """Tests using a fake filesystem."""

    def setUp(self):
        self.setUpPyfakefs(allow_root_user=False)

    def test_git_repo_info_none(self):
        """Actually not a git repo"""

        self.assertEqual(tools.get_git_repository_info(), None)

    def test_git_repo_info(self):
        """Simulate a git repo"""

        path = pathlib.Path('.git')
        path.mkdir()

        # Branch folders and hash containing file
        foobar = path / 'refs' / 'heads' / 'fix' / 'foobar'
        foobar.parent.mkdir(parents=True)

        with foobar.open('w') as handle:
            handle.write('01234')

        # HEAD file
        head = path / 'HEAD'

        with head.open('w') as handle:
            handle.write('ref: refs/heads/fix/foobar')

        # Test
        self.assertEqual(
            tools.get_git_repository_info(),
            {
                'hash': '01234',
                'branch': 'fix/foobar'
            }
        )


class ValidateSnapshotsPath(generic.TestCaseCfg):
    def test_writes(self):
        with TemporaryDirectory() as dirpath:
            ret = tools.validate_and_prepare_snapshots_path(
                path=dirpath,
                host_user_profile=self.cfg.hostUserProfile(),
                mode=self.cfg.snapshotsMode(),
                copy_links=self.cfg.copyLinks(),
                error_handler=self.cfg.notifyError)
            self.assertTrue(ret)

    def test_fails_on_ro(self):
        with TemporaryDirectory() as dirpath:
            # set directory to read only
            ro_dir_stats = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
            with generic.mockPermissions(dirpath, ro_dir_stats):
                ret = tools.validate_and_prepare_snapshots_path(
                    path=dirpath,
                    host_user_profile=self.cfg.hostUserProfile(),
                    mode=self.cfg.snapshotsMode(),
                    copy_links=self.cfg.copyLinks(),
                    error_handler=self.cfg.notifyError)
                self.assertFalse(ret)

    @patch('os.chmod')
    def test_permission_fail(self, mock_chmod):
        mock_chmod.side_effect = PermissionError()
        with TemporaryDirectory() as dirpath:
            ret = tools.validate_and_prepare_snapshots_path(
                path=dirpath,
                host_user_profile=self.cfg.hostUserProfile(),
                mode=self.cfg.snapshotsMode(),
                copy_links=self.cfg.copyLinks(),
                error_handler=self.cfg.notifyError)
            self.assertTrue(ret)


@patch(f'{tools.__name__}.datetime', wraps=datetime)
class OlderThan(unittest.TestCase):

    def test_hours_not_older(self, mock_dt):
        """Exact two hours

        Keep in mind: 20:23:00 is NOT two hours older than 18:23:00. But
        20:23:01 IS OLDER than two hours.
        """
        # year, month, day, hour=0, minute=0, second=0, microsecond=0
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)

        # exact two hours
        mock_dt.now.return_value = datetime(1982, 8, 6, 20, 23, 0, 0)

        self.assertFalse(tools.older_than(birth, 2, TimeUnit.HOUR))

    def test_hours_older(self, mock_dt):
        """Two hours plus one ms"""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)

        # two hours + 1 ms
        mock_dt.now.return_value = datetime(1982, 8, 6, 20, 23, 0, 1)

        self.assertTrue(tools.older_than(birth, 2, TimeUnit.HOUR))

    def test_days_not_older(self, mock_dt):
        """Two days"""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 8, 8, 18, 23, 0, 0)

        self.assertFalse(tools.older_than(birth, 2, TimeUnit.DAY))

    def test_days_older(self, mock_dt):
        """Two days plus one ms"""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 8, 8, 18, 23, 0, 1)

        self.assertTrue(tools.older_than(birth, 2, TimeUnit.DAY))

    def test_week_not_older(self, mock_dt):
        """Two weeks"""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 8, 20, 18, 23, 0, 0)

        self.assertFalse(tools.older_than(birth, 2, TimeUnit.WEEK))

    def test_week_older(self, mock_dt):
        """Two weeks plus one ms"""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 8, 20, 18, 23, 0, 1)

        self.assertTrue(tools.older_than(birth, 2, TimeUnit.WEEK))

    def test_month_not_older(self, mock_dt):
        """Two months."""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 10, 6, 18, 23, 0, 0)

        self.assertFalse(tools.older_than(birth, 2, TimeUnit.MONTH))

    def test_month_older(self, mock_dt):
        """Two months plus one ms."""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 10, 6, 18, 23, 0, 1)

        self.assertTrue(tools.older_than(birth, 2, TimeUnit.MONTH))

    def test_month_31th(self, mock_dt):
        """From May with 31th as last day to September with 30th as last day."""
        birth = datetime(1982, 5, 31, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 9, 30, 18, 23, 0, 0)

        self.assertFalse(tools.older_than(birth, 4, TimeUnit.MONTH))

    def test_month_31th_plus_ms(self, mock_dt):
        """Plus one ms"""
        birth = datetime(1982, 5, 31, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1982, 9, 30, 18, 23, 0, 1)

        self.assertTrue(tools.older_than(birth, 4, TimeUnit.MONTH))

    def test_month_next_year(self, mock_dt):
        """Into next year with 7 months."""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1983, 3, 6, 18, 23, 0, 0)

        self.assertFalse(tools.older_than(birth, 7, TimeUnit.MONTH))

    def test_month_next_year_plus_ms(self, mock_dt):
        """Into next year with 7 months plus 1 ms."""
        birth = datetime(1982, 8, 6, 18, 23, 0, 0)
        mock_dt.now.return_value = datetime(1983, 3, 6, 18, 23, 0, 1)

        self.assertTrue(tools.older_than(birth, 7, TimeUnit.MONTH))


class NestedDictUpdate(unittest.TestCase):
    """About nested_dict_update()"""

    def test_simple(self):
        """Simple"""
        org = {
            'foo': 7,
            'bar': {
                'a': 'drei',
                'b': 'uhr'
            }
        }
        update = {
            'planet': 'erde',
            'bar': {
                'b': 'wecker'
            },
            'ecke': 9
        }

        expect = {
            'planet': 'erde',
            'foo': 7,
            'bar': {
                'a': 'drei',
                'b': 'wecker'
            },
            'ecke': 9
        }

        self.assertDictEqual(
            tools.nested_dict_update(org, update),
            expect)



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import collections
import re
import logger


class ConfigFile:
    """Store options in a plain text file in form of: key=value
    """

    def __init__(self):
        self.dict = {}
        self.errorHandler = None
        self.questionHandler = None

    def setErrorHandler(self, handler):
        """
        Register a function that should be called for notifying errors.

        handler (method):   callable function
        """
        self.errorHandler = handler

    def setQuestionHandler(self, handler):
        """
        Register a function that should be called for asking questions.

        handler (method):   callable function
        """
        self.questionHandler = handler

    def clearHandlers(self):
        """
        Reset error and question handlers.
        """
        self.errorHandler = None
        self.questionHandler = None

    def notifyError(self, message):
        """
        Call previously registered function to show an error.

        Args:
            message (str):  error message that should be shown
        """
        if self.errorHandler is None:
            return

        self.errorHandler(message)

    def askQuestion(self, message):
        """
        Call previously registered function to ask a question.

        Args:
            message (str):    question that should be shown
        """
        if self.questionHandler is None:
            return False

        return self.questionHandler(message)

    def save(self, filename):
        """
        Save all options to file.

        Args:
            filename (str): full path

        Returns:
            bool:           ``True`` if successful
        """
        def numsort(key):
            """
            Sort int in keys in numeric order instead of alphabetical by adding
            leading zeros to int's
            """
            return re.sub(r'\d+', lambda m: m.group(0).zfill(6), key)

        try:
            with open(filename, 'wt') as f:
                keys = list(self.dict.keys())
                keys.sort(key=numsort)

                for key in keys:
                    f.write("%s=%s\n" % (key, self.dict[key]))

        except OSError as e:
            logger.error('Failed to save config: %s' % str(e), self)
            self.notifyError(
                '{}: {}'.format(_('Failed to save config'), str(e)))

            return False

        return True

    def load(self, filename, **kwargs):
        """
        Reset current options and load new options from file.

        Args:
            filename (str): full path
        """
        self.dict = {}
        self.append(filename, **kwargs)

    def append(self, filename, maxsplit=1):
        """
        Load options from file and append them to current options.

        Args:
            filename (str): full path
            maxsplit (int): split lines only n times on '='
        """
        lines = []

        if not os.path.isfile(filename):
            return

        try:
            with open(filename, 'rt') as f:
                lines = f.readlines()

        except OSError as e:
            logger.error('Failed to load config: %s' % str(e), self)
            self.notifyError(
                '{}: {}'.format(_('Failed to load config'), str(e)))

        for line in lines:
            items = line.strip('\n').split('=', maxsplit)

            if len(items) == 2:
                self.dict[items[0]] = items[1]

    def remapKey(self, old_key, new_key):
        """
        Remap keys to a new key name.

        Args:
            old_key (str):  old key name
            new_key (str):  new key name
        """
        if old_key != new_key:

            if old_key in self.dict:

                if new_key not in self.dict:
                    self.dict[new_key] = self.dict[old_key]

                del self.dict[old_key]

    def remapKeyRegex(self, pattern, replace):
        """
        Remap keys to a new key name using :py:func:`re.sub`.

        Args:
            pattern (str):  part of key name that should be replaced
            replace (:py:class:`str`, method):
                            string or a callable function which will be used
                            to replace all matches of ``pattern``.
        """
        c = re.compile(pattern)

        for key in list(self.dict):
            newKey = c.sub(replace, key)

            if key != newKey:
                self.remapKey(key, newKey)

    def hasKey(self, key):
        """
        ``True`` if key is set.

        Args:
            key (str):  string used as key

        Returns:
            bool:       ``True`` if the ``key`` is set
        """
        return key in self.dict

    def strValue(self, key, default=''):
        """
        Return a 'str' instance of key's value.

        Args:
            key (str): Key identifying the value in the config file.
            default (str): Default value if ``key`` is not present.

        Returns:
            str: Value of ``key`` or ``default``.
        """
        if key in self.dict:
            return self.dict[key]
        else:
            return default

    def setStrValue(self, key, value):
        """
        Set a string value for key.

        Args:
            key (str):      string used as key
            value (str):    store this value
        """
        self.dict[key] = value

    def intValue(self, key, default=0):
        """
        Return a 'int' instance of key's value.

        Args:
            key (str):              string used as key
            default (int):          return this if ``key`` is not set

        Returns:
            int:                    value of ``key`` or ``default``
                                    if ``key`` is not set.
        """
        try:
            return int(self.dict[key])

        except Exception:
            return default

    def setIntValue(self, key, value):
        """
        Set an integer value for key.

        Args:
            key (str):      string used as key
            value (int):    store this option
        """
        self.setStrValue(key, str(value))

    def boolValue(self, key, default=False):
        """
        Return a 'bool' instance of key's value.

        Args:
            key (str):              string used as key
            default (bool):         return this if key is not set

        Returns:
            bool:                   value of 'key' or 'default'
                                    if 'key' is not set.
        """
        try:
            val = self.dict[key]

            if "1" == val or "TRUE" == val.upper():
                return True

            return False

        except Exception:
            return default

    def setBoolValue(self, key, value):
        """
        Set a bool value for key.

        Args:
            key (str):      string used as key
            value (bool):   store this option
        """
        if value:
            self.setStrValue(key, 'true')
        else:
            self.setStrValue(key, 'false')

    def listValue(self, key, type_key='str:value', default=[]):
        """
        Return a list of values

        Size of the list must be stored in key.size

        Args:
            key (str):              used base-key
            type_key (str):         pattern of 'value-type:value-name'.
                                    See examples below.
            default (list):         default value

        Returns:
            list:                   value of ``key`` or ``default``
                                    if ``key`` is not set.

        ``type_key`` pattern examples::

            'str:value'               => return str values from key.value
            'int:type'                => return int values from key.type
            'bool:enabled'            => return bool values from key.enabled
            ('str:value', 'int:type') => return tuple of values

        """
        def typeKeySplit(tk):
            t, k = '', ''

            if isinstance(tk, str):
                t, k = tk.split(':', maxsplit=1)

            return (t, k)

        def value(key, tk):
            t, k = typeKeySplit(tk)

            if t in ('str', 'int', 'bool'):
                func = getattr(self, '%sValue' % t)

                return func('%s.%s' % (key, k))

            raise TypeError('Invalid type_key: %s' % tk)

        size = self.intValue('%s.size' % key, -1)

        if size < 0:
            return default

        ret = []

        for i in range(1, size + 1):

            if isinstance(type_key, str):

                if not self.hasKey('%s.%s.%s' %
                                   (key, i, typeKeySplit(type_key)[1])):
                    continue

                ret.append(value('%s.%s' % (key, i), type_key))

            elif isinstance(type_key, tuple):

                if not self.hasKey('%s.%s.%s' %
                                   (key, i, typeKeySplit(type_key[0])[1])):
                    continue

                items = []

                for tk in type_key:
                    items.append(value('%s.%s' % (key, i), tk))

                ret.append(tuple(items))

            else:
                raise TypeError('Invalid type_key: %s' % type_key)

        return ret

    def setListValue(self, key, type_key, value):
        """
        Set a list of values.

        Size of the list will be stored in key.size

        Args:
            key (str):      used base-key
            type_key (str): pattern of 'value-type:value-name'.
                            See examples below.
            value (list):   that should be stored

        ``type_key`` pattern examples::

            'str:value'               => return str values from key.value
            'int:type'                => return int values from key.type
            'bool:enabled'            => return bool values from key.enabled
            ('str:value', 'int:type') => return tuple of values

        """
        def setValue(key, tk, v):
            t = ''

            if isinstance(tk, str):
                t, k = tk.split(':', maxsplit=1)

            if t in ('str', 'int', 'bool'):
                func = getattr(self, 'set%sValue' % t.capitalize())

                return func('%s.%s' % (key, k), v)

            raise TypeError('Invalid type_key: %s' % tk)

        if not isinstance(value, (list, tuple)):
            raise TypeError('value has wrong type: %s' % value)

        old_size = self.intValue('%s.size' % key, -1)
        self.setIntValue('%s.size' % key, len(value))

        for i, v in enumerate(value, start=1):

            if isinstance(type_key, str):
                setValue('%s.%s' % (key, i), type_key, v)

            elif isinstance(type_key, tuple):

                for iv, tk in enumerate(type_key):

                    if len(v) > iv:
                        setValue('%s.%s' % (key, i), tk, v[iv])
                    else:
                        self.removeKey('%s.%s.%s' % (key, i, tk.split(':')[1]))
            else:
                raise TypeError('Invalid type_key: %s' % type_key)

        if len(value) < old_size:

            for i in range(len(value) + 1, old_size + 1):

                if isinstance(type_key, str):
                    self.removeKey(
                        '%s.%s.%s' % (key, i, type_key.split(':')[1]))

                elif isinstance(type_key, tuple):
                    for tk in type_key:
                        self.removeKey(
                            '%s.%s.%s' % (key, i, tk.split(':')[1]))

    def removeKey(self, key):
        """
        Remove key from options.

        Args:
            key (str):    string used as key
        """
        if key in self.dict:
            del self.dict[key]

    def removeKeysStartsWith(self, prefix):
        """
        Remove key from options which start with given prefix.

        Args:
            prefix (str):   prefix for keys (key starts with this string)
                            that should be removed
        """
        removeKeys = []

        for key in self.dict.keys():
            if key.startswith(prefix):
                removeKeys.append(key)

        for key in removeKeys:
            del self.dict[key]

    def keys(self):
        return list(self.dict.keys())


class ConfigFileWithProfiles(ConfigFile):
    """
    Store options in profiles as 'profileX.key=value'

    Args:
        default_profile_name (str): default name of the first profile.
    """

    def __init__(self, default_profile_name=''):
        ConfigFile.__init__(self)

        self.default_profile_name = default_profile_name
        self.current_profile_id = '1'
        self.setCurrentProfile(self.current_profile_id)

    def load(self, filename):
        """
        Reset current options and load new options from file.

        Args:
            filename (str): full path
        """
        self.current_profile_id = '1'
        super(ConfigFileWithProfiles, self).load(filename)

    def append(self, filename):
        """
        Load options from file and append them to current options.

        Args:
            filename (str): full path
        """
        super(ConfigFileWithProfiles, self).append(filename)

        found = False
        profiles = self.profiles()

        for profile_id in profiles:

            if profile_id == self.current_profile_id:
                found = True
                break

        if not found and profiles:
            self.current_profile_id = profiles[0]

        if self.intValue('profiles.version') <= 0:
            rename_keys = []

            for key in self.dict.keys():
                if key.startswith('profile.0.'):
                    rename_keys.append(key)

            for old_key in rename_keys:
                new_key = 'profile1.' + old_key[10:]
                self.dict[new_key] = self.dict[old_key]
                del self.dict[old_key]

        if self.intValue('profiles.version') != 1:
            self.setIntValue('profiles.version', 1)

    def profiles(self):
        """
        List of all available profile IDs. Profile IDs are strings!

        Returns:
            list: List with strings of profile IDs as strings.
        """
        return self.strValue(key='profiles', default='1').split(':')

    def profilesSortedByName(self):
        """
        List of available profile IDs alphabetically sorted by their names.
        Profile IDs are strings!

        Returns:
            list:   all available profile IDs as strings
        """
        profiles_unsorted = self.profiles()

        if len(profiles_unsorted) <= 1:
            return profiles_unsorted

        profiles_dict = {}

        for profile_id in profiles_unsorted:
            profiles_dict[self.profileName(profile_id).upper()] = profile_id

        # sort the dictionary by key (the profile name)
        profiles_sorted = collections.OrderedDict(
            sorted(profiles_dict.items()))

        # return the names as a list
        return list(profiles_sorted.values())

    def currentProfile(self):
        """
        Currently selected profile ID. Profile IDs are strings!

        Returns:
            str:    profile ID
        """
        return self.current_profile_id

    def setCurrentProfile(self, profile_id):
        """
        Change the current profile.

        Args:
            profile_id (str, int):  valid profile ID

        Returns:
            bool:                   ``True`` if successful
        """
        if isinstance(profile_id, int):
            profile_id = str(profile_id)

        profiles = self.profiles()

        for i in profiles:

            if i == profile_id:

                profile_name = self.profileName(profile_id)

                self.current_profile_id = profile_id
                logger.changeProfile(profile_id, profile_name)
                logger.debug(
                    f'Change current profile: {profile_id}={profile_name}',
                    self)

                return True

        return False

    def setCurrentProfileByName(self, name):
        """
        Change the current profile by a given name.

        Args:
            name (str): valid profile name

        Returns:
            bool: ``True`` if successful
        """

        # Find the profile_id to this name...
        for profile_id in self.profiles():
            if self.profileName(profile_id) == name:
                # ...and set current profile by this id.
                return self.setCurrentProfile(profile_id)

        return False

    def profileExists(self, profile_id):
        """
        ``True`` if the profile exists.

        Args:
            profile_id (str, int):  profile ID

        Returns:
            bool:                   ``True`` if ``profile_id`` exists.
        """
        if isinstance(profile_id, int):
            profile_id = str(profile_id)

        return profile_id in self.profiles()

    def profileExistsByName(self, name):
        """
        ``True`` if the profile exists.

        Args:
            name (str): profile name

        Returns:
            bool:       ``True`` if ``name`` exists.
        """
        profiles = self.profiles()

        for profile_id in profiles:

            if self.profileName(profile_id) == name:
                return True

        return False

    def profileName(self, profile_id=None):
        """Name of the profile.

        Args:
            profile_id (str, int): Valid profile ID

        Returns:
            str: Name of profile.
        """
        if isinstance(profile_id, int):
            profile_id = str(profile_id)

        if profile_id is None:
            profile_id = self.current_profile_id

        if profile_id == '1':
            default = self.default_profile_name
        else:
            default = 'Profile %s' % profile_id

        return self.profileStrValue('name', default, profile_id)

    def addProfile(self, name):
        """
        Add a new profile if the name is not already in use.

        Args:
            name (str): new profile name

        Returns:
            str:        new profile ID
        """
        profiles = self.profiles()

        for profile_id in profiles:

            if self.profileName(profile_id) == name:
                self.notifyError(_(
                    'Profile "{name}" already exists.').format(name=name))

                return None

        new_id = 1

        while True:
            ok = True

            if str(new_id) in profiles:
                ok = False

            if ok:
                break

            new_id = new_id + 1

        new_id = str(new_id)

        profiles.append(new_id)
        self.setStrValue('profiles', ':'.join(profiles))

        self.setProfileStrValue('name', name, new_id)

        return new_id

    def removeProfile(self, profile_id=None):
        """
        Remove profile and all its keys and values.

        Args:
            profile_id (str, int):  valid profile ID

        Returns:
            bool:   ``True`` if successful
        """
        if isinstance(profile_id, int):
            profile_id = str(profile_id)

        if profile_id is None:
            profile_id = self.current_profile_id

        profiles = self.profiles()

        if len(profiles) <= 1:
            self.notifyError(_("The last profile cannot be removed."))

            return False

        found = False
        index = 0

        for profile in profiles:

            if profile == profile_id:
                self.removeKeysStartsWith(self.profileKey('', profile_id))

                del profiles[index]

                self.setStrValue('profiles', ':'.join(profiles))
                found = True

                break

            index = index + 1

        if not found:
            return False

        if self.current_profile_id == profile_id:
            self.current_profile_id = '1'

        return True

    def setProfileName(self, name, profile_id=None):
        """
        Change the name of the profile.

        Args:
            name (str):             new profile name
            profile_id (str, int):  valid profile ID

        Returns:
            bool:                   ``True`` if successful.
        """
        if isinstance(profile_id, int):
            profile_id = str(profile_id)

        if profile_id is None:
            profile_id = self.current_profile_id

        profiles = self.profiles()

        for profile in profiles:

            if self.profileName(profile) == name:

                if profile[0] != profile_id:
                    self.notifyError(_(
                        'Profile "{name}" already exists.').format(name=name))

                    return False

        self.setProfileStrValue('name', name, profile_id)

        return True

    def profileKey(self, key, profile_id=None):
        """
        Prefix for keys with profile. e.g. 'profile1.key'

        Args:
            key (str): Key identifier.
            profile_id (str, int): Valid profile ID.

        Returns:
            str: Key with prefix 'profile1.key'
        """
        if isinstance(profile_id, int):
            profile_id = str(profile_id)

        if profile_id is None:
            profile_id = self.current_profile_id

        return 'profile' + profile_id + '.' + key

    def removeProfileKey(self, key, profile_id=None):
        """
        Remove the key from profile.

        Args:
            key (str):              key name
            profile_id (str, int):  valid profile ID
        """
        self.removeKey(self.profileKey(key, profile_id))

    def removeProfileKeysStartsWith(self, prefix, profile_id=None):
        """
        Remove the keys starting with prefix from profile.

        Args:
            prefix (str):           prefix for keys (key starts with this
                                    string) that should be removed.
            profile_id (str, int):  valid profile ID
        """
        self.removeKeysStartsWith(self.profileKey(prefix, profile_id))

    def remapProfileKey(self, oldKey, newKey, profileId=None):
        """
        Remap profile keys to a new key name.

        Args:
            oldKey (str):           old key name
            newKey (str):           new key name
            profileId (str, int):   valid profile ID
        """
        self.remapKey(self.profileKey(oldKey, profileId),
                      self.profileKey(newKey, profileId))

    def hasProfileKey(self, key, profile_id=None):
        """
        ``True`` if key is set in profile.

        Args:
            key (str):              string used as key
            profile_id (str, int):  valid profile ID

        Returns:
            bool:                   ``True`` if ``key`` is set.
        """
        return self.profileKey(key, profile_id) in self.dict

    def profileStrValue(self, key, default='', profile_id=None):
        """Return the value of ``key`` related to ``profile_id``.

        Returns:
           str: The value.
        """
        return self.strValue(self.profileKey(key, profile_id), default)

    def setProfileStrValue(self, key, value, profile_id=None):
        self.setStrValue(self.profileKey(key, profile_id), value)

    def profileIntValue(self, key, default=0, profile_id=None):
        return self.intValue(self.profileKey(key, profile_id), default)

    def setProfileIntValue(self, key, value, profile_id=None):
        self.setIntValue(self.profileKey(key, profile_id), value)

    def profileBoolValue(self, key, default=False, profile_id=None):
        return self.boolValue(self.profileKey(key, profile_id), default)

    def setProfileBoolValue(self, key, value, profile_id=None):
        self.setBoolValue(self.profileKey(key, profile_id), value)

    def profileListValue(self,
                         key,
                         type_key='str:value',
                         default=[],
                         profile_id=None):
        return self.listValue(
            self.profileKey(key, profile_id), type_key, default)

    def setProfileListValue(self, key, type_key, value, profile_id=None):
        self.setListValue(self.profileKey(key, profile_id), type_key, value)



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2008-2022 Taylor Raak
# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""The widget to setup scheduling backup jobs."""
import datetime
from PyQt6.QtWidgets import (QHBoxLayout,
                             QFormLayout,
                             QGroupBox,
                             QWidget,
                             QLabel,
                             QLineEdit,
                             QSpinBox,
                             QCheckBox)
import config
import tools
import qttools
from manageprofiles import combobox


class ScheduleWidget(QGroupBox):
    """Widget about schedule snapshots.

    That widget is used in the 'General' tab of the 'Manage profiles' dialog.
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, parent):
        super().__init__(title=_('Schedule'), parent=parent)

        main_layout = QFormLayout(self)

        def _create_form_entry(label: str, widget: QWidget = None) -> int:
            """Helper to create a row with a label and widget in the form
            layout.

            The returned index is used later to toggle visibility of that
            row.
            """
            if widget:
                main_layout.addRow(label, widget)
            else:
                lbl = QLabel(label)
                lbl.setWordWrap(True)
                main_layout.addRow(lbl)

            return main_layout.rowCount() - 1

        # Schedule modes
        self._combo_schedule_mode = self._schedule_mode_combobox()
        main_layout.addRow(self._combo_schedule_mode)

        # Day
        self._combo_day = self._day_combobox()
        self._rowidx_day = _create_form_entry(_('Day:'), self._combo_day)

        # Weekday
        self._combo_weekday = self._weekday_combobox()
        self._rowidx_weekday = _create_form_entry(
            _('Weekday:'), self._combo_weekday)

        # Time (formerly known as Hour)
        self._combo_time = self._time_combobox()
        self._rowidx_time = _create_form_entry(
            _('Time:'), self._combo_time)

        # HourS
        self._edit_cronpattern = QLineEdit(self)
        self._rowidx_cronpattern = _create_form_entry(
            _('Hours:'), self._edit_cronpattern)

        # Offset
        self._spin_offset = QSpinBox(self)
        self._spin_offset.setSingleStep(1)
        self._spin_offset.setRange(0, 59)
        hlayout = QHBoxLayout()
        hlayout.addWidget(self._spin_offset)
        hlayout.addWidget(QLabel(_('after the hour'), self))
        hlayout.addStretch()
        self._rowidx_offset = _create_form_entry(_('Minutes:'), hlayout)

        # Udev
        self._rowidx_udev = _create_form_entry(
            _('Run Back In Time as soon as the drive is connected (only once'
              ' every X days). You will be prompted for your sudo password.'))

        # Repeatedly (like anacron)
        self._rowidx_repeated = _create_form_entry(
            _('Run Back In Time repeatedly. This is useful if the computer '
              'is not running regularly.'))

        # Repeatedly - Every (value) (units)
        self._spin_period = QSpinBox(self)
        self._spin_period.setSingleStep(1)
        self._spin_period.setRange(1, 10000)
        self._combo_repeated_unit = self._repeated_unit_combobox()
        hlayout = QHBoxLayout()
        hlayout.addWidget(self._spin_period)
        hlayout.addWidget(self._combo_repeated_unit)
        hlayout.addStretch()
        self._rowidx_period = _create_form_entry(_('Every:'), hlayout)

        # Debug logging
        self._check_debug = QCheckBox(self)
        self._check_debug.setText(_('Enable logging of debug messages'))
        qttools.set_wrapped_tooltip(
            self._check_debug,
            [
                _('Writes debug-level messages into the system log via '
                  '"--debug".'),
                _('Caution: Only use this temporarily for diagnostics, as it '
                  'generates a large amount of output.')
            ]
        )
        self._rowidx_debug = main_layout.rowCount()
        main_layout.addRow(self._check_debug)

        # Signal
        self._combo_schedule_mode.currentIndexChanged.connect(
            self._slot_schedule_mode_changed)
        self._slot_schedule_mode_changed(None)

    def _schedule_mode_combobox(self) -> combobox.BitComboBox:
        """Create the the combobox for schedule mode.

        Returns:
            BitComboBox: The widget.
        """
        # pylint: disable=protected-access

        # Regular schedule modes for that combo box
        schedule_modes = {
            config.Config.NONE: _('Disabled'),
            config.Config.AT_EVERY_BOOT: _('At every boot/reboot'),
            config.Config._5_MIN: ngettext(
                'Every {n} minute', 'Every {n} minutes', 5).format(n=5),
            config.Config._10_MIN: ngettext(
                'Every {n} minute', 'Every {n} minutes', 10).format(n=10),
            config.Config._30_MIN: ngettext(
                'Every {n} minute', 'Every {n} minutes', 30).format(n=30),
            config.Config._1_HOUR: ngettext(
                'Every hour', 'Every {n} hours', 1).format(n=1),
            config.Config._2_HOURS: ngettext(
                'Every {n} hour', 'Every {n} hours', 2).format(n=2),
            config.Config._4_HOURS: ngettext(
                'Every {n} hour', 'Every {n} hours', 4).format(n=4),
            config.Config._6_HOURS: ngettext(
                'Every {n} hour', 'Every {n} hours', 6).format(n=6),
            config.Config._12_HOURS: ngettext(
                'Every {n} hour', 'Every {n} hours', 12).format(n=12),
            config.Config.CUSTOM_HOUR: _('Custom hours'),
            config.Config.DAY: _('Every day'),
            config.Config.REPEATEDLY: _('Repeatedly (anacron)'),
            config.Config.UDEV: _('When drive gets connected (udev)'),
            config.Config.WEEK: _('Every week'),
            config.Config.MONTH: _('Every month'),
            config.Config.YEAR: _('Every year')
        }

        return combobox.BitComboBox(self, schedule_modes)

    def _time_combobox(self) -> combobox.BitComboBox:
        """Combobox with time/hours (e.g. 03:00).

        Returns:
            BitComboBox: The widget.
        """

        # Dev note (buhtz): strftime is needed because of localization
        # {0: '00:00', 100: '01:00', ..., 2200: '22:00', 2300: '23:00'}
        times = {
            val*100: datetime.time(val, 0).strftime('%H:%M')
            for val in range(0, 24)
        }
        return combobox.BitComboBox(self, times)

    def _day_combobox(self) -> combobox.BitComboBox:
        """Combobox with day number in the month.

        Returns:
            BitComboBox: The widget.
        """
        days = {day: str(day) for day in range(1, 29)}
        return combobox.BitComboBox(self, days)

    def _weekday_combobox(self) -> combobox.BitComboBox:
        """Combobox with name of the weekday.

        Returns:
            BitComboBox: The widget.
        """
        sunday = datetime.date(2011, 11, 6)
        weekdays = {
            day: (sunday + datetime.timedelta(days=day)).strftime('%A')
            for day in range(1, 8)
        }
        return combobox.BitComboBox(self, weekdays)

    def _repeated_unit_combobox(self):
        """Combobox for "Every ..." schedule mode to select the units to use.

        Returns:
            BitComboBox: The widget.
        """
        repeatedly_units = {
            config.Config.HOUR: _('Hour(s)'),
            config.Config.DAY: _('Day(s)'),
            config.Config.WEEK: _('Week(s)'),
            config.Config.MONTH: _('Month(s)')
        }

        return combobox.BitComboBox(self, repeatedly_units)

    def _slot_schedule_mode_changed(self, _idx):
        """Handle value changed events for schedule mode combobox."""
        self._set_child_visibilities(self._combo_schedule_mode.current_data)

    def _set_child_visibilities(self, backup_mode_id: int):
        """Modify the visibility of child widgets (addressed by their index in
        the form layout) based on the selected schedule mode.
        """
        layout = self.layout()

        layout.setRowVisible(
            self._rowidx_cronpattern,
            backup_mode_id == config.Config.CUSTOM_HOUR)

        layout.setRowVisible(
            self._rowidx_weekday,
            backup_mode_id == config.Config.WEEK)

        layout.setRowVisible(
            self._rowidx_day,
            backup_mode_id == config.Config.MONTH)

        layout.setRowVisible(
            self._rowidx_time,
            backup_mode_id >= config.Config.DAY)

        layout.setRowVisible(
            self._rowidx_offset,
            backup_mode_id in config.Config.HOURLY_BACKUPS
        )

        vis = config.Config.REPEATEDLY <= backup_mode_id <= config.Config.UDEV
        layout.setRowVisible(
            self._rowidx_period,
            vis)

        if vis is True:
            layout.setRowVisible(self._rowidx_time, False)

        layout.setRowVisible(
            self._rowidx_repeated,
            backup_mode_id == config.Config.REPEATEDLY)

        layout.setRowVisible(
            self._rowidx_udev,
            backup_mode_id == config.Config.UDEV)

    def load_values(self, cfg: config.Config):
        """Set the values of the widgets regarding the current config."""

        self._combo_schedule_mode.select_by_data(cfg.scheduleMode())

        self._combo_time.select_by_data(cfg.scheduleTime())
        self._combo_day.select_by_data(cfg.scheduleDay())
        self._combo_weekday.select_by_data(cfg.scheduleWeekday())
        self._spin_offset.setValue(cfg.schedule_offset())

        self._edit_cronpattern.setText(cfg.customBackupTime())

        self._spin_period.setValue(cfg.scheduleRepeatedPeriod())
        self._combo_repeated_unit.select_by_data(cfg.scheduleRepeatedUnit())

        self._check_debug.setChecked(cfg.scheduleDebug())

        self._slot_schedule_mode_changed(
            self._combo_schedule_mode.currentIndex())

    def store_values(self, cfg: config.Config) -> bool:
        """Store the widgets values in the config instance.

        Args:
            cfg: The configuration data instance.

        Returns:
            bool: Success or not.
        """

        if self._combo_schedule_mode.current_data == cfg.CUSTOM_HOUR:
            # Dev note (buhtz, 2024-05): IMHO checkCronPattern() is not needed
            # because the "crontab" command itself will validate this. See
            # schedule.write_crontab().
            # We just need to take care to catch an the error in the GUI
            # and report it to the user.
            # An alternative solution would be a GUI element where the user
            # is not able to input an invalid value. See #1449 about redesign
            # the schedule section in the Manage Profiles dialog.
            if not tools.checkCronPattern(self._edit_cronpattern.text()):

                cfg.errorHandler(
                    _('Custom hours can only be a comma separated list of '
                      'hours (e.g. 8,12,18,23) or */3 for periodic '
                      'backups every 3 hours.')
                )

                return False

        cfg.setScheduleMode(self._combo_schedule_mode.current_data)
        cfg.setScheduleTime(self._combo_time.current_data)

        if cfg.scheduleMode() in config.Config.HOURLY_BACKUPS:
            cfg.set_schedule_offset(self._spin_offset.value())
        else:
            cfg.set_schedule_offset(config.Config.DEFAULT_OFFSET)

        cfg.setScheduleWeekday(self._combo_weekday.current_data)
        cfg.setScheduleDay(self._combo_day.current_data)
        cfg.setCustomBackupTime(self._edit_cronpattern.text())
        cfg.setScheduleRepeatedPeriod(self._spin_period.value())
        cfg.setScheduleRepeatedUnit(self._combo_repeated_unit.current_data)
        cfg.setScheduleDebug(self._check_debug.isChecked())

        return True



# SPDX-FileCopyrightText: © 2007 Sander Marechal
# SPDX-FileCopyrightText: © 2016 Germar Reitze
#
# SPDX-License-Identifier: CC0-1.0
#
# This file is released under Creative Commons Zero 1.0 (CC0-1.0) and part of
# the program "Back In Time". The program as a whole is released under GNU
# General Public License v2 or any later version (GPL-2.0-or-later).
# See file/folder LICENSE or
# go to <https://spdx.org/licenses/CC0-1.0.html>
# and <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""A generic daemon class.

    Original from:
    http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python
    Copyright © 2007 Sander Marechal
    License CC0 or Public Domain

Notes about the license (by buhtz, 2024-08-13):

    The linked blog article is licensed under CC BY-SA 3.0 which is not
    compatible with GPLv2 used by Back In Time. But the original author
    clarified that the code used in the blog article is public domain.

See this original email.

    Date: Tue, 13 Aug 2024 14:19:48 +0200
    Subject: Re: Your Daemon code in Back In Time
    Message-ID: <06084b4b-2293-4a28-a290-96fa4d309a8b@email.android.com>
    In-Reply-To: <3d57067b590e271ce6f361ce4147ac08@posteo.de>
    From: Sander Marechal <sander@marechal.io>
    To: c.buhtz@posteo.jp

    Hello Christian,

    As far as I am concerned that daemon code is public domain. You can use it
    under any license you want. There are only a few ways to start a daemon so
    technically the code can't be copyrighted as far as I am concerned. That CC
    license is just for my articles in general.

    Kind regards,

    --
    Sander Marechal
"""
import sys
import os
import signal
import atexit
import errno
from time import sleep
import logger
from applicationinstance import ApplicationInstance


def fdDup(old, new_fd, mode='w'):
    """Duplicate file descriptor `old` to `new_fd` and closing the latter
    first.

    Used to redirect stdin, stdout and stderr from daemonized threads.

    Args:
        old (str): Path to the old file (e.g. /dev/stdout).
        new_fd (_io.TextIOWrapper): File object for the new file.
        mode (str): Mode in which the old file should be opened.
    """
    try:
        fd = open(old, mode)
        os.dup2(fd.fileno(), new_fd.fileno())

    except OSError as e:
        logger.debug('Failed to redirect {}: {}'.format(old, str(e)))


class Daemon:
    """A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """
    def __init__(self,
                 pidfile=None,
                 stdin='/dev/null',
                 stdout='/dev/stdout',
                 stderr='/dev/null',
                 umask=0o022):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile
        self.umask = umask
        if pidfile:
            self.appInstance = ApplicationInstance(
                pidfile, autoExit=False, flock=False)

    def daemonize(self):
        """Converts the current process into a daemon.

        It is a process running in the background, sending a SIGTERM signal to
        the current process. This is done via the UNIX double-fork magic, see
        Stevens' 'Advanced Programming in the UNIX Environment' for details
        (ISBN 0201563177) and this explanation:
        https://stackoverflow.com/a/6011298
        """
        try:
            pid = os.fork()
            logger.debug('first fork pid: {}'.format(pid), self)

            if pid > 0:
                # exit first parent
                sys.exit(0)

        except OSError as e:
            logger.error("fork #1 failed: %d (%s)" % (e.errno, str(e)), self)
            sys.exit(1)

        # Decouple from parent environment
        logger.debug('decouple from parent environment', self)
        os.chdir("/")
        os.setsid()
        os.umask(self.umask)

        # Do second fork
        try:
            pid = os.fork()
            logger.debug('second fork pid: {}'.format(pid), self)

            if pid > 0:
                # exit from second parent
                sys.exit(0)

        except OSError as e:
            logger.error("fork #2 failed: %d (%s)" % (e.errno, str(e)), self)
            sys.exit(1)

        # redirect standard file descriptors
        logger.debug('redirect standard file descriptors', self)

        sys.stdout.flush()
        sys.stderr.flush()
        fdDup(self.stdin, sys.stdin, 'r')
        fdDup(self.stdout, sys.stdout, 'w')
        fdDup(self.stderr, sys.stderr, 'w')

        signal.signal(signal.SIGTERM, self.cleanupHandler)

        if self.pidfile:
            atexit.register(self.appInstance.exitApplication)

            # write pidfile
            logger.debug('write pidfile', self)
            self.appInstance.startApplication()

    def cleanupHandler(self, signum, frame):
        if self.pidfile:
            self.appInstance.exitApplication()
        sys.exit(0)

    def start(self):
        """Start the daemon."""

        # Check for a pidfile to see if the daemon already runs
        if self.pidfile and not self.appInstance.check():
            logger.error(f'pidfile {self.pidfile} already exists. '
                         'Daemon already running?\n', self)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """Stop the daemon."""
        if not self.pidfile:
            logger.debug(
                "Unattended daemon can't be stopped. No PID file", self)
            return

        # Get the pid from the pidfile
        pid = self.appInstance.readPidFile()[0]

        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            logger.error(message % self.pidfile, self)
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while True:
                os.kill(pid, signal.SIGTERM)
                sleep(0.1)

        except OSError as err:

            if err.errno == errno.ESRCH:
                # No such process
                self.appInstance.exitApplication()

            else:
                logger.error(str(err), self)
                sys.exit(1)

    def restart(self):
        """Restart the daemon."""
        self.stop()
        self.start()

    def reload(self):
        """Send SIGHUP signal to process."""
        if not self.pidfile:
            logger.debug(
                "Unattended daemon can't be reloaded. No PID file", self)
            return

        # Get the pid from the pidfile
        pid = self.appInstance.readPidFile()[0]

        if not pid:
            logger.error(f'pidfile {self.pidfile} does not exist. '
                         'Daemon not running?\n', self)
            return

        # Try killing the daemon process
        try:
            os.kill(pid, signal.SIGHUP)

        except OSError as err:

            if err.errno == errno.ESRCH:
                # no such process
                self.appInstance.exitApplication()

            else:
                sys.stderr.write(str(err))
                sys.exit(1)

    def status(self):
        """Return status."""
        if not self.pidfile:
            logger.debug(
                "Unattended daemon can't be checked. No PID file", self)
            return

        return not self.appInstance.check()

    def run(self):
        """Override this method when subclass ``Daemon``. It will be called
        after the process has been daemonized by ``start()`` or ``restart()``.
        """
        pass
